# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""This file contains the Grid protocol defining the minimal arrays contained in a grid"""

from typing import Protocol

from power_grid_model_ds._core.model.arrays import (
    BranchArray,
    NodeArray,
    ThreeWindingTransformerArray,
)


class MinimalGridArrays(Protocol):
    """Protocol for the minimal arrays contained in a grid,
    they may be implemented using properties or added as attributes"""

    node: NodeArray
    three_winding_transformer: ThreeWindingTransformerArray
    branches: BranchArray
    branch_arrays: list[BranchArray]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Misc utils"""

from collections.abc import Sequence
from typing import Type, get_type_hints

import numpy as np


def is_sequence(seq):
    """
    Returns True for lists, tuples, sets, arrays
    Return False for strings, dicts
    """
    if isinstance(seq, str):
        return False

    if isinstance(seq, (np.ndarray, set)):
        return True
    return isinstance(seq, Sequence)


def get_inherited_attrs(cls: Type, *private_attributes):
    """
    Get the attribute from the object and all its parents
    """

    # The extras are needed for annotated types like NDArray3
    retrieved_attributes = get_type_hints(cls, include_extras=True)
    retrieved_attributes = {attr: type for attr, type in retrieved_attributes.items() if not attr.startswith("_")}

    for private_attr in private_attributes:
        for parent in reversed(list(cls.__mro__)):
            attr_dict = retrieved_attributes.get(private_attr, {})
            attr_dict.update(getattr(parent, private_attr, {}))
            retrieved_attributes[private_attr] = attr_dict

    return retrieved_attributes



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Stores the GraphContainer class"""

import dataclasses
from dataclasses import dataclass
from typing import Generator

import numpy as np

from power_grid_model_ds._core.model.arrays import Branch3Array, BranchArray, NodeArray
from power_grid_model_ds._core.model.arrays.base.array import FancyArray
from power_grid_model_ds._core.model.arrays.base.errors import RecordDoesNotExist
from power_grid_model_ds._core.model.containers.grid_protocol import MinimalGridArrays
from power_grid_model_ds._core.model.graphs.models import RustworkxGraphModel
from power_grid_model_ds._core.model.graphs.models.base import BaseGraphModel


@dataclass
class GraphContainer:
    """Contains graphs"""

    active_graph: BaseGraphModel
    """The graph containing only active branches."""

    complete_graph: BaseGraphModel
    """The graph containing all branches."""

    @property
    def graph_attributes(self) -> Generator:
        """Get all graph attributes of the container.

        Yield:
            BaseGraphModel: The next graph attribute.
        """
        return (field for field in dataclasses.fields(self) if isinstance(getattr(self, field.name), BaseGraphModel))

    @classmethod
    def empty(cls, graph_model: type[BaseGraphModel] = RustworkxGraphModel) -> "GraphContainer":
        """Get empty instance of GraphContainer.

        Args:
            graph_model (type[BaseGraphModel]): The graph model to use. Defaults to RustworkxGraphModel.
              An alternative graph model can be passed as an argument.

        Returns:
            GraphContainer: The empty graph container.
        """

        return cls(
            active_graph=graph_model(active_only=True),
            complete_graph=graph_model(active_only=False),
        )

    def add_branch(self, branch: BranchArray) -> None:
        """Add a branch to all graphs"""
        for field in self.graph_attributes:
            graph = getattr(self, field.name)
            graph.add_branch_array(branch_array=branch)
            setattr(self, field.name, graph)

    def add_branch3(self, branch: Branch3Array) -> None:
        """Add a branch to all graphs"""
        for field in self.graph_attributes:
            graph = getattr(self, field.name)
            graph.add_branch3_array(branch3_array=branch)
            setattr(self, field.name, graph)

    def delete_branch(self, branch: BranchArray) -> None:
        """Remove a branch from all graphs"""
        for field in self.graph_attributes:
            graph = getattr(self, field.name)
            graph.delete_branch_array(branch_array=branch)
            setattr(self, field.name, graph)

    def delete_branch3(self, branch: Branch3Array) -> None:
        """Remove a branch from all graphs"""
        for field in self.graph_attributes:
            graph = getattr(self, field.name)
            graph.delete_branch3_array(branch3_array=branch)
            setattr(self, field.name, graph)

    def add_node(self, node: NodeArray) -> None:
        """Add a node to all graphs"""
        for field in dataclasses.fields(self):
            graph = getattr(self, field.name)
            graph.add_node_array(node_array=node, raise_on_fail=False)
            setattr(self, field.name, graph)

    def delete_node(self, node: NodeArray) -> None:
        """Remove a node from all graphs"""
        for field in dataclasses.fields(self):
            graph = getattr(self, field.name)
            graph.delete_node_array(node_array=node)
            setattr(self, field.name, graph)

    def make_active(self, branch: BranchArray) -> None:
        """Add branch to all active_only graphs"""

        from_node = branch.from_node.item()
        to_node = branch.to_node.item()
        for field in dataclasses.fields(self):
            graph = getattr(self, field.name)
            if graph.active_only:
                graph.add_branch(from_ext_node_id=from_node, to_ext_node_id=to_node)
            setattr(self, field.name, graph)

    def make_inactive(self, branch: BranchArray) -> None:
        """Remove a branch from all active_only graphs"""

        from_node = branch.from_node.item()
        to_node = branch.to_node.item()
        for field in dataclasses.fields(self):
            graph = getattr(self, field.name)
            if graph.active_only:
                graph.delete_branch(from_ext_node_id=from_node, to_ext_node_id=to_node)
            setattr(self, field.name, graph)

    @classmethod
    def from_arrays(cls, arrays: MinimalGridArrays) -> "GraphContainer":
        """Build from arrays"""
        cls._validate_branches(arrays=arrays)

        new_container = cls.empty()
        for graph_field in new_container.graph_attributes:
            graph = getattr(new_container, graph_field.name)
            new_graph = graph.from_arrays(arrays, active_only=graph.active_only)
            setattr(new_container, graph_field.name, new_graph)

        return new_container

    @staticmethod
    def _validate_branches(arrays: MinimalGridArrays):
        for array in arrays.branch_arrays:
            if any(~np.isin(array.from_node, arrays.node.id)):
                raise RecordDoesNotExist(f"Found invalid .from_node values in {array.__class__.__name__}")
            if any(~np.isin(array.to_node, arrays.node.id)):
                raise RecordDoesNotExist(f"Found invalid .to_node values in {array.__class__.__name__}")

    def _append(self, array: FancyArray) -> None:
        if isinstance(array, BranchArray):
            self.add_branch(array)
        if isinstance(array, Branch3Array):
            self.add_branch3(array)
        if isinstance(array, NodeArray):
            self.add_node(array)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from power_grid_model_ds._core.model.constants import EMPTY_ID, empty

__all__ = ["EMPTY_ID", "empty"]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import numpy as np
from numpy.typing import NDArray

from power_grid_model_ds._core.model.dtypes.id import Id


class Regulator(Id):
    "Regulator data type"

    regulated_object: NDArray[np.int32]  # a valid regulated object ID
    status: NDArray[np.int8]  # connection status of regulated object


class TransformerTapRegulator(Regulator):
    """Transformer tap regulator data type"""

    control_side: NDArray[np.int8]  # the controlled side of the transformer (see BranchSide/Branch3Side of PGM)
    u_set: NDArray[np.float64]  # the voltage setpoint
    u_band: NDArray[np.float64]  # the width of the voltage band
    line_drop_compensation_r: NDArray[np.float64]  # compensation for voltage drop due to resistance during transport
    line_drop_compensation_x: NDArray[np.float64]  # compensation for voltage drop due to reactance during transport

    _defaults = {
        "line_drop_compensation_r": 0,
        "line_drop_compensation_x": 0,
    }



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Create a grid from text a text file"""

import logging
from typing import TYPE_CHECKING

from power_grid_model_ds._core.model.enums.nodes import NodeType

if TYPE_CHECKING:
    from power_grid_model_ds._core.model.grids.base import Grid

_logger = logging.getLogger(__name__)


class TextSource:
    """Class for handling text sources.

    Text sources are only intended for test purposes so that a grid can quickly be designed from a text file.
    Moreover, these text sources are compatible with the grid editor at https://csacademy.com/app/graph_editor/

    Example of a text file:
        S1 2
        2 3
        3 4 transformer
        4 5
        S1 7

    See docs/examples/3_drawing_a_grid.md for more information.
    """

    def __init__(self, grid_class: type["Grid"]):
        self.grid = grid_class.empty()

    def load_from_txt(self, *args: str) -> "Grid":
        """Load a grid from text"""

        text_lines = [line for arg in args for line in arg.strip().split("\n")]

        txt_nodes, txt_branches = self.read_txt(text_lines)
        self.add_nodes(txt_nodes)
        self.add_branches(txt_branches)
        self.grid.set_feeder_ids()
        return self.grid

    @staticmethod
    def read_txt(txt_lines: list[str]) -> tuple[set, dict]:
        """Extract assets from text"""

        txt_nodes = set()
        txt_branches = {}
        for text_line in txt_lines:
            if not text_line.strip() or text_line.startswith("#"):
                continue  # skip empty lines and comments
            try:
                from_node_str, to_node_str, *comments = text_line.strip().split()
            except ValueError as err:
                raise ValueError(f"Text line '{text_line}' is invalid. Skipping...") from err
            comments = comments[0].split(",") if comments else []

            txt_nodes |= {from_node_str, to_node_str}
            txt_branches[(from_node_str, to_node_str)] = comments
        return txt_nodes, txt_branches

    def add_nodes(self, nodes: set[str]):
        """Add nodes to the grid"""
        source_nodes = {int(node[1:]) for node in nodes if node.startswith("S")}
        regular_nodes = {int(node) for node in nodes if not node.startswith("S")}

        if source_nodes.intersection(regular_nodes):
            raise ValueError("Source nodes and regular nodes have overlapping ids")

        for node_id in source_nodes:
            new_node = self.grid.node.empty(1)
            new_node.id = node_id
            new_node.node_type = NodeType.SUBSTATION_NODE
            self.grid.append(new_node, check_max_id=False)

        for node_id in regular_nodes:
            new_node = self.grid.node.empty(1)
            new_node.id = node_id
            self.grid.append(new_node, check_max_id=False)

    def add_branches(self, branches: dict[tuple[str, str], list[str]]):
        """Add branches to the grid"""
        for branch, comments in branches.items():
            self.add_branch(branch, comments)

    def add_branch(self, branch: tuple[str, str], comments: list[str]):
        """Add a branch to the grid"""
        from_node_str, to_node_str = branch
        from_node = int(from_node_str.replace("S", ""))
        to_node = int(to_node_str.replace("S", ""))

        if "transformer" in comments:
            new_branch = self.grid.transformer.empty(1)
        elif "link" in comments:
            new_branch = self.grid.link.empty(1)
        else:  # assume it is a line
            new_branch = self.grid.line.empty(1)

        branch_ids = [branch_id for branch_id in comments if branch_id.isdigit()]
        if branch_ids:
            if len(branch_ids) > 1:
                raise ValueError(f"Multiple branch ids found in row {branch} {','.join(comments)}")
            new_branch.id = int(branch_ids[0])

        new_branch.from_node = from_node
        new_branch.to_node = to_node
        new_branch.from_status = 1
        if "open" in comments:
            new_branch.to_status = 0
        else:
            new_branch.to_status = 1
        self.grid.append(new_branch, check_max_id=False)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""helper functions for pickling python objects and loading pickle objects"""

import pickle
from pathlib import Path

from power_grid_model_ds._core.utils.zip import gzip2file


def save_to_pickle(path: Path, python_object: object):
    """Save a python object to pickle"""
    path.parent.mkdir(exist_ok=True, parents=True)
    with open(str(path), "wb") as file:
        pickle.dump(python_object, file)


def load_from_pickle(path: Path) -> object:
    """Load a python object from a pickle file"""
    with open(str(path), "rb") as file:
        return pickle.load(file)


def get_pickle_path(path: Path) -> Path:
    """
    Returns the path to the pickle file.
    If ony a .gz-file is available, the .gz-file is unpacked.
    """
    pickle_suffix = ".pickle"
    gz_suffix = ".gz"

    if path.suffix == pickle_suffix and path.is_file():
        return path

    if path.with_suffix(pickle_suffix).is_file():
        return path.with_suffix(pickle_suffix)

    if path.suffix == gz_suffix and path.is_file():
        return gzip2file(path)

    pickle_gz_suffix = pickle_suffix + gz_suffix
    if path.with_suffix(pickle_gz_suffix).is_file():
        return gzip2file(path.with_suffix(pickle_gz_suffix))

    raise FileNotFoundError(f"Expected either {path.name}.pickle or {path.name}.pickle.gz")



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from power_grid_model_ds._core.fancypy import array_equal, concatenate, sort, unique
from power_grid_model_ds._core.model.arrays.base.array import FancyArray
from power_grid_model_ds._core.model.containers.base import FancyArrayContainer

__all__ = ["FancyArray", "FancyArrayContainer", "concatenate", "unique", "sort", "array_equal"]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from power_grid_model_ds._core.load_flow import PGMCoreException
from power_grid_model_ds._core.model.arrays.base.errors import (
    ArrayDefinitionError,
    MultipleRecordsReturned,
    RecordDoesNotExist,
)
from power_grid_model_ds._core.model.graphs.errors import (
    GraphError,
    MissingBranchError,
    MissingNodeError,
    NoPathBetweenNodes,
)

__all__ = [
    "PGMCoreException",
    "GraphError",
    "ArrayDefinitionError",
    "RecordDoesNotExist",
    "MultipleRecordsReturned",
    "MissingNodeError",
    "MissingBranchError",
    "NoPathBetweenNodes",
]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Generator for LineArray"""

import numpy as np

from power_grid_model_ds._core import fancypy as fp
from power_grid_model_ds._core.data_source.generator.arrays.base import BaseGenerator
from power_grid_model_ds._core.model.arrays import LineArray, TransformerArray
from power_grid_model_ds._core.model.grids.base import Grid

AVERAGE_ROUTE_SIZE = 20


class LineGenerator(BaseGenerator):
    """Generator for line elements in the grid"""

    def __init__(self, grid: Grid, seed: int) -> None:
        super().__init__(grid=grid, seed=seed)
        self.connected_nodes: list = []
        self.unconnected_nodes: list = []
        self.line_array: LineArray = self.grid.line.__class__()
        self.trafo_array: TransformerArray = self.grid.transformer.__class__()

    # pylint: disable=arguments-differ
    def run(self, amount: int, number_of_routes: int | None = None) -> LineArray:
        """Generate routes, lines and normally open points (NOPs)"""

        self.trafo_array = self.grid.transformer
        if number_of_routes is None:
            number_of_routes = self.determine_number_of_routes()
        if number_of_routes > 0:
            self.create_routes(number_of_routes)
        else:
            self.line_array = self.grid.line

        # while not all connected, add lines from a connected node to an unconnected node
        self.set_unconnected_nodes()
        while any(self.unconnected_nodes):
            self.connect_nodes()
            self.set_unconnected_nodes()

        number_of_nops = amount
        if number_of_nops > 0:
            self.create_nop_lines(number_of_nops)

        return self.line_array

    def create_routes(self, number_of_routes: int):
        """Create a number of lines from the substation to unconnected nodes"""
        # each source should have at least one route
        number_of_sources = len(self.grid.source)
        from_nodes = self.rng.choice(self.grid.source.node, number_of_routes - number_of_sources, replace=True)
        not_source_mask = ~np.isin(self.grid.node.id, self.grid.source.node)
        to_nodes = self.rng.choice(self.grid.node.id[not_source_mask], number_of_routes, replace=False)
        capacities = 100 + self.rng.exponential(200, number_of_routes)
        line_array = self.grid.line.__class__.zeros(number_of_routes)
        line_array.id = 1 + self.grid.max_id + np.arange(number_of_routes)
        line_array.from_node = np.concatenate((self.grid.source.node, from_nodes))
        line_array.to_node = to_nodes
        line_array.from_status = [1] * number_of_routes
        line_array.to_status = [1] * number_of_routes
        line_array.r1 = self.rng.exponential(0.2, number_of_routes)
        line_array.x1 = self.rng.exponential(0.02, number_of_routes)
        line_array.i_n = capacities
        self.line_array = line_array

    def determine_number_of_routes(self) -> int:
        """Decide on a number of routes based on expected route-size"""
        expected_number_of_routes = int(np.ceil(len(self.grid.node) / AVERAGE_ROUTE_SIZE))
        number_of_sources = len(self.grid.source)
        # The number of routes is the max of the number of sources
        # and the expected number based on size
        return max(expected_number_of_routes, number_of_sources)

    def connect_nodes(self):
        """Add a new line between an active and inactive line"""
        to_node = self.rng.choice(self.unconnected_nodes)
        to_voltage = self.grid.node[self.grid.node.id == to_node].u_rated[0]
        same_voltage_mask = self.grid.node.u_rated == to_voltage
        same_voltage_nodes = self.grid.node[same_voltage_mask]
        options_mask = np.isin(self.connected_nodes, same_voltage_nodes.id)
        from_node = self.rng.choice(np.array(self.connected_nodes)[options_mask])
        capacity = 100 + self.rng.exponential(200, 1)
        new_line = self.grid.line.__class__.zeros(1)
        new_line.id = 1 + max(max(self.line_array.id), self.grid.max_id)  # pylint: disable=nested-min-max
        new_line.from_node = from_node
        new_line.to_node = to_node
        new_line.from_status = [1]
        new_line.to_status = [1]
        new_line.r1 = self.rng.exponential(0.2, 1)
        new_line.x1 = self.rng.exponential(0.02, 1)
        new_line.i_n = capacity
        self.line_array = fp.concatenate(self.line_array, new_line)

    def create_nop_lines(self, number_of_nops: int):
        """Create the inactive lines between different routes (Normally Open Points)"""
        nops = [self.rng.choice(self.grid.node.id, 2, replace=False) for _ in range(number_of_nops)]
        from_nodes = [nop[0] for nop in nops]
        to_nodes = [nop[1] for nop in nops]
        capacities = 100 + self.rng.exponential(200, number_of_nops)
        nop_lines = self.grid.line.__class__.zeros(number_of_nops)
        nop_lines.id = 1 + self.line_array.id.max() + np.arange(number_of_nops)
        nop_lines.from_node = from_nodes
        nop_lines.to_node = to_nodes
        nop_lines.from_status = [1] * number_of_nops
        nop_lines.to_status = [0] * number_of_nops
        nop_lines.r1 = self.rng.exponential(0.2, number_of_nops)
        nop_lines.x1 = self.rng.exponential(0.02, number_of_nops)
        nop_lines.i_n = capacities
        self.line_array = fp.concatenate(self.line_array, nop_lines)

    def set_unconnected_nodes(self) -> None:
        """From a line array and total set of nodes determine which are not yet connected"""
        connected_link_mask = np.logical_or(
            np.isin(self.grid.node.id, self.line_array.from_node),
            np.isin(self.grid.node.id, self.line_array.to_node),
        )
        connected_trafo_mask = np.logical_or(
            np.isin(self.grid.node.id, self.trafo_array.from_node),
            np.isin(self.grid.node.id, self.trafo_array.to_node),
        )
        connected_mask = np.logical_or(
            connected_link_mask,
            connected_trafo_mask,
        )
        connected_nodes = self.grid.node.id[connected_mask]
        unconnected_nodes = self.grid.node.id[~connected_mask]

        self.unconnected_nodes = unconnected_nodes.tolist()
        self.connected_nodes = connected_nodes.tolist()



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Stores the FancyArrayContainer class"""

import dataclasses
import inspect
import logging
from dataclasses import dataclass
from typing import Type, TypeVar

import numpy as np

from power_grid_model_ds._core import fancypy as fp
from power_grid_model_ds._core.model.arrays.base.array import FancyArray
from power_grid_model_ds._core.model.arrays.base.errors import RecordDoesNotExist
from power_grid_model_ds._core.model.constants import EMPTY_ID

Self = TypeVar("Self", bound="FancyArrayContainer")


@dataclass
class FancyArrayContainer:
    """
    Base class for ArrayContainers.
    Contains general functionality that is nonspecific to the type of array being stored.
    """

    _id_counter: int

    @property
    def id_counter(self):
        """Returns the private _id_counter field (as read-only)"""
        return self._id_counter

    @classmethod
    def empty(cls: Type[Self]) -> Self:
        """Create an empty grid"""
        empty_fields = cls._get_empty_fields()
        return cls(**empty_fields)

    def all_arrays(self):
        """Returns all arrays in the container."""

        for field in dataclasses.fields(self):
            attribute = getattr(self, field.name)
            if isinstance(attribute, FancyArray):
                yield attribute

    @classmethod
    def find_array_field(cls, array_type: Type[FancyArray]) -> dataclasses.Field:
        """Find the Field that holds an array of type array_type.

        Args:
            array_type(type[FancyArray]): FancyArray subclass.

        Raises:
            TypeError: if no field with the given `array_type` is found or if multiple fields are found.

        Returns:
            a Field instance.
        """
        fields = [
            field
            for field in dataclasses.fields(cls)
            if inspect.isclass(field.type) and issubclass(field.type, array_type)
        ]
        if (nr_fields := len(fields)) != 1:
            raise TypeError(
                f"Expected to find 1 array with type '{array_type.__name__}' in {cls.__name__} ({nr_fields} found)"
            )
        return fields[0]

    @property
    def max_id(self) -> int:
        """Returns the max id across all arrays within the container."""
        max_per_array = [np.max(array.id) if array.size > 0 else 0 for array in self.all_arrays()]
        return int(max(max_per_array))

    def check_ids(self, check_between_arrays: bool = True, check_within_arrays: bool = True) -> None:
        """Checks for duplicate id values across all arrays in the container.

        Args:
            check_between_arrays(bool): whether to check for duplicate ids across arrays
            check_within_arrays(bool): whether to check for duplicate ids within each array

        Raises:
            ValueError: if duplicates are found.
        """

        id_arrays = [array for array in self.all_arrays() if hasattr(array, "id")]
        if not id_arrays:
            return  # no arrays to check

        duplicates_between_arrays = self._get_duplicates_between_arrays(
            id_arrays, check=check_between_arrays
        )  # if check_between_arrays else []
        arrays_with_duplicates = self._get_arrays_with_duplicates(id_arrays, check=check_within_arrays)

        if not any(duplicates_between_arrays) and not any(arrays_with_duplicates):
            return

        if any(duplicates_between_arrays):
            logging.warning(f"The following ids occur in multiple arrays: {duplicates_between_arrays}!")
        for array_class in arrays_with_duplicates:
            logging.warning(f"{array_class.__name__} contains duplicates!")

        raise ValueError(f"Duplicates found within {self.__class__.__name__}!")

    def append(self, array: FancyArray, check_max_id: bool = True) -> None:
        """Append the given asset_array to the corresponding field of ArrayContainer and generate ids.

        Args:
            array(FancyArray): the asset_array to be appended (e.g. a NodeArray instance).
            check_max_id(bool): whether to check max(array.id) with the id counter

        Returns:
            None
        """
        self._append(array=array, check_max_id=check_max_id)

    def attach_ids(self, array: FancyArray) -> FancyArray:
        """Generate and attach ids to the given FancyArray. Also updates _id_counter.

        Args:
            array(FancyArray): the array of which the id column is set.

        Returns:
            FancyArray: initial array with updated `id` column.
        """
        if not array.size:
            return array

        if (id_set := set(array.id)) != {array.get_empty_value("id")}:
            raise ValueError(f"Cannot attach ids to array that contains non-empty ids: {id_set}")

        start = self._id_counter + 1
        end = start + len(array)
        array.id = np.arange(start, end)
        self._id_counter = max(self._id_counter, end - 1)

        return array

    def search_for_id(self, record_id: int) -> list[FancyArray]:
        """Attempts to find a record across all id-arrays within the container.

        This method is only intended for debugging purposes since it is very inefficient.
        In normal circumstances you should use ``get`` or ``filter`` to find records within a specific array.

        Args:
            record_id(int): the id of the record to be found.

        Returns:
         list[FancyArray]:a list of arrays that contain the given record_id.
         Each array within the list contains all records with the given array.
        """

        logging.warning("Using search_for_id(). Make sure to use only while debugging!")

        arrays_with_record = []

        id_arrays = [array for array in self.all_arrays() if "id" in array.dtype.names]
        for id_array in id_arrays:
            matching_records = id_array.filter(id=record_id)
            if matching_records.size:
                arrays_with_record.append(matching_records)

        if arrays_with_record:
            return arrays_with_record
        raise RecordDoesNotExist(f"record id '{record_id}' not found in {self.__class__.__name__}")

    def _append(self, array: FancyArray, check_max_id: bool = True) -> None:
        """
        Append the given asset_array to the corresponding field of Grid and generate ids.
        Args:
            array: the asset_array to be appended (e.g. a KabelArray instance).
            check_max_id: whether to check max(array.id) with the id counter
        Returns: None.
        """
        if array.size == 0:
            return

        array_field = self.find_array_field(array.__class__)

        if hasattr(array, "id"):
            self._update_id_counter(array, check_max_id)

        # Add the given asset_array to the corresponding array in the Grid.
        array_attr = getattr(self, array_field.name)
        appended = fp.concatenate(array_attr, array)
        setattr(self, array_field.name, appended)

    @classmethod
    def _get_empty_fields(cls) -> dict:
        empty_fields = {}

        empty_fields.update(cls._get_empty_arrays())
        empty_fields.update({"_id_counter": 0})
        return empty_fields

    @classmethod
    def _get_empty_arrays(cls) -> dict:
        return {
            field.name: field.type()
            for field in dataclasses.fields(cls)
            if inspect.isclass(field.type) and issubclass(field.type, FancyArray)
        }

    def _update_id_counter(self, array, check_max_id: bool = True):
        if np.all(array.id == EMPTY_ID):
            array = self.attach_ids(array)
        elif np.any(array.id == EMPTY_ID):
            raise ValueError(f"Cannot append: array contains empty [{EMPTY_ID}] and non-empty ids.")

        new_max_id = np.max(array.id)
        if check_max_id and new_max_id < self._id_counter:
            raise ValueError(f"Cannot append: id {new_max_id} is lower than the id counter")

        # Update _id_counter
        self._id_counter = max(self._id_counter, new_max_id)

    @staticmethod
    def _get_duplicates_between_arrays(id_arrays: list[FancyArray], check: bool) -> np.ndarray:
        if not check:
            return np.array([])
        unique_ids_per_array = [np.unique(array.id) for array in id_arrays]

        all_ids = np.concatenate(unique_ids_per_array)

        unique_ids, counts = np.unique(all_ids, return_counts=True)
        duplicate_mask = counts > 1
        return unique_ids[duplicate_mask]

    @staticmethod
    def _get_arrays_with_duplicates(id_arrays: list[FancyArray], check: bool) -> list:
        arrays_with_duplicates: list[Type] = []
        if not check:
            return arrays_with_duplicates
        for id_array in id_arrays:
            duplicates: np.ndarray = id_array.check_ids(return_duplicates=True)
            if duplicates.size > 0:
                arrays_with_duplicates.append(id_arrays.__class__)
        return arrays_with_duplicates



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from power_grid_model_ds._core.model.enums.nodes import NodeType

__all__ = ["NodeType"]
